
class Permission:
    """Permission Object
    """

    def __init__(self, data) -> None:
        self.perms = {
            "CREATE_INSTANT_INVITE": 0x1,
            "KICK_MEMBERS": 0x2,
            "BAN_MEMBERS": 0x4,
            "ADMINISTRATOR" : 0x8,
            "MANAGE_CHANNELS": 0x10,
            "MANAGE_GUILD" : 0x20,
            "ADD_REACTIONS" : 0X40,
            "VIEW_AUDIT_LOG" : 0x80,
            "PRIORITY_SPEAKER" : 0x100,
            "STREAM" : 0x200,
            "VIEW_CHANNEL" : 0x400,
            "SEND_MESSAGES" : 0x800,
            "SEND_TTS_MESSAGES" : 0x1000,
            "MANAGE_MESSAGES" : 0x2000,
            "EMBED_LINKS" : 0x4000,
            "ATTACH_FILES" : 0x8000,
            "READ_MESSAGE_HISTORY" : 0x10000,
            "MENTION_EVERYONE" : 0x20000,
            "USE_EXTERNAL_EMOJIS" : 0x40000,
            "VIEW_GUILD_INSIGHTS" : 0x80000,
            "CONNECT" : 0x100000,
            "SPEAK" : 0x200000,
            "MUTE_MEMBERS" : 0x400000,
            "DEAFEN_MEMBERS" : 0x800000,
            "MOVE_MEMBERS" : 0x1000000,
            "USE_VAD" : 0x2000000,
            "CHANGE_NICK" : 0x4000000,
            "MANAGE_NICK" : 0x8000000,
            "MANAGE_ROLES" : 0x10000000,
            "MANAGE_WEBHOOKS" : 0x20000000,
            "MANAGE_EMOJIS_AND_STICKERS" : 0x40000000,
            "USE_APPLICATION_COMMANDS" : 0x80000000,
            "REQUEST_TO_SPEAK" : 0x100000000,
            "MANAGE_EVENTS" : 0x200000000,
            "MANAGE_THREADS" : 0x400000000,
            "CREATE_PUBLIC_THREADS" : 0x800000000,
            "CREATE_PRIVATE_THREADS" : 0x1000000000,
            "USE_EXTERNAL_STICKERS" : 0x2000000000,
            "SEND_MESSAGES_IN_THREADS" : 0x4000000000,
            "USE_EMBEDDED_ACTIVITIES" : 0x8000000000,
            "MODERATE_MEMBERS" : 0x10000000000
        }
        self.permissions = self.calculate_permissions(data)


    def calculate_permissions(self, perm_value):
        permissions = []
        for perm, value in self.perms.items():
            if (perm_value & value) == value:
                permissions.append(perm)

        return permissions






class Webhook:
    def __init__(self, data: dict, bot, http) -> None:
        self.http = http
        self.bot = bot
        self._update(data)

    def _update(self, data):
        self.id = data.get("id")
        self.type = data.get("type")
        self.guild_id = data.get("guild_id")
        self.channel_id = data.get("channel_id")
        self.name = data.get("name")
        self.avatar = data.get("avatar")
        self.token = data.get("token")
        self.application_id = data.get("application_id")
        self.webhook_url = f"https://discord.com/api/webhooks/{self.id}/{self.token}"
        self.source_guild = data.get("source_guild")
        self.source_channel = data.get("source_channel")

    async def send(self, content):
        await self.http.request(method = "post", endpoint = f"/webhooks/{self.id}/{self.token}", json = {"content": content})

    async def delete(self):
        await self.http.request(method = "delete", endpoint = f"/webhooks/{self.id}/{self.token}")



"""Data objects from discord, used throughout selfcord
"""
from .user import User
from .client import Client
from .guild import Guild
from .channel import TextChannel, Category, VoiceChannel, DMChannel, GroupChannel
from .message import Message
from .emoji import Emoji
from .permission import Permission
from .role import Role
from .webhook import Webhook
from .member import Member



from aiohttp import ClientSession
from aioconsole import aprint
import asyncio
from base64 import b64encode
import aiohttp
from selfcord.api.errors import LoginFailure
import random
from ..models import User, Client

class http:
    def __init__(self) -> None:
        xproperties = ['eyJvcyI6IldpbmRvd3MiLCJicm93c2VyIjoiRmlyZWZveCIsImRldmljZSI6IiIsInN5c3RlbV9sb2NhbGUiOiJmciIsImJyb3dzZXJfdXNlcl9hZ2VudCI6Ik1vemlsbGEvNS4wIChXaW5kb3dzIE5UIDEwLjA7IFdpbjY0OyB4NjQ7IHJ2OjEwMi4wKSBHZWNrby8yMDEwMDEwMSBGaXJlZm94LzEwMi4wIiwiYnJvd3Nlcl92ZXJzaW9uIjoiMTAyLjAiLCJvc192ZXJzaW9uIjoiMTAiLCJyZWZlcnJlciI6IiIsInJlZmVycmluZ19kb21haW4iOiIiLCJyZWZlcnJlcl9jdXJyZW50IjoiIiwicmVmZXJyaW5nX2RvbWFpbl9jdXJyZW50IjoiIiwicmVsZWFzZV9jaGFubmVsIjoic3RhYmxlIiwiY2xpZW50X2J1aWxkX251bWJlciI6MTU0MTg2LCJjbGllbnRfZXZlbnRfc291cmNlIjpudWxsfQ==', 'eyJvcyI6IkxpbnV4IiwiYnJvd3NlciI6IkRpc2NvcmQgQ2xpZW50IiwicmVsZWFzZV9jaGFubmVsIjoiY2FuYXJ5IiwiY2xpZW50X3ZlcnNpb24iOiIwLjAuMTQwIiwib3NfdmVyc2lvbiI6IjUuMTkuMC0zLXJ0MTAtTUFOSkFSTyIsIm9zX2FyY2giOiJ4NjQiLCJzeXN0ZW1fbG9jYWxlIjoiZW4tR0IiLCJ3aW5kb3dfbWFuYWdlciI6IktERSx1bmtub3duIiwiZGlzdHJvIjoiXCJNYW5qYXJvIExpbnV4XCIiLCJjbGllbnRfYnVpbGRfbnVtYmVyIjoxNTQyMTYsImNsaWVudF9ldmVudF9zb3VyY2UiOm51bGx9', 'eyJvcyI6IkxpbnV4IiwiYnJvd3NlciI6IkNocm9tZSIsImRldmljZSI6IiIsInN5c3RlbV9sb2NhbGUiOiJlbi1HQiIsImJyb3dzZXJfdXNlcl9hZ2VudCI6Ik1vemlsbGEvNS4wIChYMTE7IExpbnV4IHg4Nl82NCkgQXBwbGVXZWJLaXQvNTM3LjM2IChLSFRNTCwgbGlrZSBHZWNrbykgQ2hyb21lLzEwNi4wLjAuMCBTYWZhcmkvNTM3LjM2IiwiYnJvd3Nlcl92ZXJzaW9uIjoiMTA2LjAuMC4wIiwib3NfdmVyc2lvbiI6IiIsInJlZmVycmVyIjoiaHR0cHM6Ly93d3cucm9ibG94LmNvbS8iLCJyZWZlcnJpbmdfZG9tYWluIjoid3d3LnJvYmxveC5jb20iLCJyZWZlcnJlcl9jdXJyZW50IjoiIiwicmVmZXJyaW5nX2RvbWFpbl9jdXJyZW50IjoiIiwicmVsZWFzZV9jaGFubmVsIjoic3RhYmxlIiwiY2xpZW50X2J1aWxkX251bWJlciI6MTU0MTg2LCJjbGllbnRfZXZlbnRfc291cmNlIjpudWxsfQ==']

        self.cookies        = {}
        self.token          = None
        self.xproperties    = random.choice(xproperties)
        self.base_url       = 'https://discord.com/api/v9'

    async def static_login(self, token: str):
        '''Used to retrieve basic token information

        Args:
            token (str): User token

        Returns:
            Client: A Client object
        '''
        await self.get_cookie()
        self.token = token
        data = await self.request('get', '/users/@me')
        self.client = Client(data)
        return data

    async def get_cookie(self):
        async with aiohttp.ClientSession() as session:
            async with session.get('https://discord.com', headers={'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) discord/0.0.139 Chrome/91.0.4472.164 Electron/13.6.6 Safari/537.36'}) as resp:
                dcf = resp.headers['set-cookie'].split('__dcfduid=')[0].split(';')[0]
                sdc = resp.headers['set-cookie'].split('__sdcfduid=')[0].split(';')[0]
                cfr = resp.headers['set-cookie'].split('__cfruid=')[0].split(';')[0]

                self.cookies['dcf'] = self.cookies.get('dcf') if dcf != "" else ""
                self.cookies['sdc'] = self.cookies.get('sdc') if sdc != "" else ""
                self.cookies['cfr'] = self.cookies.get('cfr') if cfr != "" else ""
                self.cookie = set(self.cookies)

            async with session.get('https://discord.com/api/v9/experiments', headers={'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) discord/0.0.139 Chrome/91.0.4472.164 Electron/13.6.6 Safari/537.36'}) as resp:
                json = await resp.json()
                self.fingerprint = json['fingerprint']

    def remove_dupes(self, dictionary: dict):
        return set(dictionary)

    async def request(self, method: str, endpoint: str, *args, **kwargs) -> dict:
        '''Used to send requests

        Args:
            method (str): HTTP method
            endpoint (str): Discord api endpoint

        Raises:
            LoginFailure: If you suck

        Returns:
            dict: Data, json data
        '''
        url = self.base_url + endpoint

        headers = {
            'cookie': f'{self.cookie}',
            'authorization': self.token,
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) discord/0.0.139 Chrome/91.0.4472.164 Electron/13.6.6 Safari/537.36',
            'Content-Type': 'application/json',
            'X-Super-Properties': self.xproperties,
            'X-Discord-Locale': 'en-GB',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-origin',
            'origin': 'https://discord.com',
            'x-debug-options': 'logGatewayEvents,logOverlayEvents,logAnalyticsEvents,bugReporterEnabled',
            'x-fingerprint': self.fingerprint,
            'TE': 'trailers',
        }

        async with ClientSession(headers = headers) as session:
            request = getattr(session, method)
            while True:
                async with request(url, *args, **kwargs) as resp:
                    if resp.status == 429:
                        try:
                            json = await resp.json()
                            await asyncio.sleep(json['retry_after'])
                            continue
                        except Exception as e:
                            await aprint(f'Error: {e}')
                            text = await resp.text()
                            await aprint(text)
                            break

                    elif resp.status == 401:
                        json = await resp.json()
                        raise LoginFailure(json, resp.status)

                    elif resp.status == 403:
                        json = await resp.json()
                        await aprint(json)

                    elif resp.status == 201:
                        data = await resp.json()
                        break

                    elif resp.status == 204:
                        data = await resp.text()
                        break

                    elif resp.ok:
                        data = await resp.json()
                        break

                    else:
                        json = await resp.json()
                        raise LoginFailure(json, resp.status)
        try:
            if resp.headers['set-cookie']:
                dcf = resp.headers['set-cookie'].split('__dcfduid=')[0].split(';')[0]
                sdc = resp.headers['set-cookie'].split('__sdcfduid=')[0].split(';')[0]
                cfr = resp.headers['set-cookie'].split('__cfruid=')[0].split(';')[0]
                bm = resp.headers['set-cookie'].split('__cf_bm=')[0].split(';')[0]

                self.cookies['dcf'] = self.cookies.get('dcf') if dcf != "" else ""
                self.cookies['sdc'] = self.cookies.get('sdc') if sdc != "" else ""
                self.cookies['cfr'] = self.cookies.get('cfr') if cfr != "" else ""
                self.cookies['bm'] = self.cookies.get('bm') if bm != "" else ""
                self.cookie = set(self.cookies)
        except: pass

        return data

    async def encode_image(self, url):
        async with ClientSession() as session:
            async with session.get(f'{url}') as resp:
                image = b64encode(await resp.read())
                newobj = str(image).split('"', 2)

        return f'data:image/png;base64,{newobj[1]}'




from .message import Message
import time
from .user import User
from .webhook import Webhook
import asyncio

from selfcord.models import message


class TextChannel:
    """
    Text Channel Object
        Represents a Guild/Server channel within Discord.
        All methods are coroutines and thus need to be awaited.

    Returns:
        Name of the channel

    Functions:
        - delete()
        - edit(name[str], parent_id[int], position[int], topic[str])
        - history()
        - purge(amount[int])
        - spam(amount[int], content[str], tts[bool]=False)
        - send()
        - reply()
        - create_webhook()

    Example Object (ref: https://discord.com/developers/docs/resources/channel#channel-object-example-guild-text-channel)
        {
          "id": "41771983423143937",
          "guild_id": "41771983423143937",
          "name": "general",
          "type": 0,
          "position": 6,
          "permission_overwrites": [],
          "rate_limit_per_user": 2,
          "nsfw": true,
          "topic": "24/7 chat about how to gank Mike #2",
          "last_message_id": "155117677105512449",
          "parent_id": "399942396007890945",
          "default_auto_archive_duration": 60
        }
    """

    def __init__(self, data, bot, http) -> None:
        self.permissions = []
        self.webhooks = []
        self.http = http
        self.bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def _update(self, data):
        self.topic = data.get("topic")
        self.rate_limit_per_user = data.get("rate_limit_per_user")
        self.position = data.get("position")
        self.name = data.get("name")
        self.id = data.get("id")
        self.guild_id = data.get("guild_id")
        self.last_message_id = data.get("last_message_id")
        self.flags = data.get("flags")
        self.default_thread_rate_limit_per_user = data.get("default_thread_rate_limit_per_user")
        self.category_id = data.get("parent_id")

    async def delete(self):
        """
        Deletes the text channel object.

        Args:
            No arguments required

        Returns:
            No return value
        """
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}")
        del self

    async def edit(self, name: str = None, parent_id: int = None, position: int = None, topic: str = None):
        """
        Edits the text channel object details. Requires the `Manage Channels` permission.
        Not all details can be modified.

        Args:
            - name(str) : Optional - Specifies a new name for the channel object. Defaults to None.
            - parent_id(int) : Optional - Specifies a new parent (category) for the text channel object.
            - position(int) : Optional - Modifies the sorting position of the text channel within the guild.
            - topic(str) : Optional - Modifies the channel topic (0-4096 characters for GUILD_FORUM channels,
                                        0-1024 characters for all others)
        Notice: Each parent category can contain up to 50 channels.

        Returns:
            e(str) : Exception parsed as string value.
        """
        payload = {}
        if name is not None:
            payload['name'] = name
        if parent_id is not None:
            payload['parent_id'] = parent_id
        if position is not None:
            payload['position'] = position
        if topic is not None and topic != "":
            payload['topic'] = topic

        try:
            await self.http.request(method="patch", endpoint=f"/channels/{self.id}", json=payload)
        except Exception as e:
            return str(e)

    async def history(self):
        """
        Get channel message history.

        Args:
            No arguments required

        Returns:
            messages(list) : List of messages from the channel.
        """
        messages = []
        data = await self.http.request(method="get", endpoint=f"/channels/{self.id}/messages?limit=100")
        for msg in data:
            messages.append(Message(msg, self.bot, self.http))
        while True:
            data = await self.http.request(method="get",
                                           endpoint=f"/channels/{self.id}/messages?limit=100&before={data[-1]['id']}")
            if len(data) > 0:
                for msg in data:
                    messages.append(Message(msg, self.bot, self.http))
            else:
                break

        return messages

    async def purge(self, amount: int = None):
        """
        Delete a number of messages, starting from the most recent.

        Args:
            amount(int) : Number of messages to purge/delete.

        Returns:
            No return value
        """
        messages = await self.history()
        msgs = []
        for msg in messages:
            if str(msg.author.id) == str(self.bot.user.id):
                msgs.append(msg)

        if amount != None:
            for i in range(0, len(msgs[:amount]), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[:amount][i:i + 3]))
        else:
            for i in range(0, len(msgs), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[i:i + 3]))

    async def spam(self, amount: int, content: str, tts=False):
        """
        Send multiple of the same message.

        Args:
            - amount(int) : Number of spam messages to send.
            - content(str) : The message to send.
            - tts(bool) = False : Specify whether it is a TTS message.

        Returns:
             No return value.
        """
        amount: list[int] = [i + 1 for i in range(amount)]
        for i in range(0, len(amount), 3):
            await asyncio.gather(
                *(asyncio.create_task(self.send(tts=tts, content=content)) for amoun in amount[i:i + 3]))

    async def send(self, content=None, tts=False):
        """
        Send a message to the text channel.

        Args:
            - content(str) : Message content. Should be string type or similar. Discord `embed` type is not allowed.
            - tts(bool) :
        """
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/{self.guild_id}/{self.id}"},
                                json={"content": content, "tts": tts})

    async def reply(self, message, content=None, tts=False):
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/{self.guild_id}/{self.id}"},
                                json={"content": content, "tts": tts,
                                      "message_reference": {"channel_id": f"{self.id}", "message_id": f"{message.id}"},
                                      "allowed_mentions": {"parse": ["users", "roles", "everyone"],
                                                           "replied_user": False}})

    async def create_webhook(self, name: str = None, avatar_url: str = None):
        fields = {}
        if name != None:
            fields['name'] = name
        else:
            raise TypeError("Name is required...")
        if avatar_url != None:
            data = await self.http.encode_image(avatar_url)
            fields['avatar'] = data
        data = await self.http.request(method="post", endpoint=f"/channels/{self.id}/webhooks", json=fields)
        self.webhooks.append(Webhook(data, self.http))


class VoiceChannel:
    """Voice Channel Object
    """

    def __init__(self, data, bot, http) -> None:
        self.permissions = []
        self.webhooks = []
        self.http = http
        self.bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def _update(self, data):
        self.name = data.get("name")
        self.id = data.get("id")
        self.guild_id = data.get("guild_id")
        self.last_message_id = data.get("last_message_id")
        self.rtc_region = data.get("rtc_region")
        self.flags = data.get("flags")
        self.bitrate = data.get("bitrate")
        self.rate_limit_per_user = data.get("rate_limit_per_user")
        self.position = data.get("position")
        self.category_id = data.get("parent_id")

    async def delete(self):
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}")
        del self

    async def history(self):
        messages = []
        data = await self.http.request(method="get", endpoint=f"/channels/{self.id}/messages?limit=100")
        for msg in data:
            messages.append(Message(msg, self.bot, self.http))
        while True:
            data = await self.http.request(method="get",
                                           endpoint=f"/channels/{self.id}/messages?limit=100&before={data[-1]['id']}")
            if len(data) > 0:
                for msg in data:
                    messages.append(Message(msg, self.bot, self.http))
            else:
                break

        return messages

    async def purge(self, amount: int = None):
        messages = await self.history()
        msgs = []
        for msg in messages:
            if str(msg.author.id) == str(self.bot.user.id):
                msgs.append(msg)

        if amount != None:
            for i in range(0, len(msgs[:amount]), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[:amount][i:i + 3]))
        else:
            for i in range(0, len(msgs), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[i:i + 3]))

    async def spam(self, amount: int, content: str, tts=False):
        amount: list[int] = [i + 1 for i in range(amount)]
        for i in range(0, len(amount), 3):
            await asyncio.gather(
                *(asyncio.create_task(self.send(tts=tts, content=content)) for amoun in amount[i:i + 3]))

    async def send(self, content=None, tts=False):
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/{self.guild_id}/{self.id}"},
                                json={"content": content, "tts": tts})

    async def reply(self, message, content=None, tts=False):
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/{self.guild_id}/{self.id}"},
                                json={"content": content, "tts": tts,
                                      "message_reference": {"channel_id": f"{self.id}", "message_id": f"{message.id}"},
                                      "allowed_mentions": {"parse": ["users", "roles", "everyone"],
                                                           "replied_user": False}})

    async def create_webhook(self, name: str = None, avatar_url: str = None):
        fields = {}
        if name != None:
            fields['name'] = name
        else:
            raise TypeError("Name is required...")
        if avatar_url != None:
            data = await self.http.encode_image(avatar_url)
            fields['avatar'] = data
        data = await self.http.request(method="post", endpoint=f"/channels/{self.id}/webhooks", json=fields)
        self.webhooks.append(Webhook(data, self.http))


class Category:
    """Category Object
    """

    def __init__(self, data, bot, http) -> None:
        self.bot = bot
        self.http = http
        self.permissions = []
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def _update(self, data):
        self.name = data.get("name")
        self.id = data.get("id")
        self.guild_id = data.get("guild_id")
        self.position = data.get("position")
        self.flags = data.get("flags")

    async def delete(self):
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}")
        del self


class DMChannel:
    """DM Channel Object
    """

    def __init__(self, data, bot, http) -> None:
        self.http = http
        self.bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.recipient}"

    def _update(self, data):
        self.recipient = User(data.get("recipients")[0], self.bot, self.http)
        self.last_message_id = data.get("last_message_id")
        self.id = data.get("id")
        self.flags = data.get("id")

    async def delete(self):
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}?silent=false")
        del self

    async def history(self):
        messages = []
        data = await self.http.request(method="get", endpoint=f"/channels/{self.id}/messages?limit=100")
        for msg in data:
            messages.append(Message(msg, self.bot, self.http))
        while True:
            data = await self.http.request(method="get",
                                           endpoint=f"/channels/{self.id}/messages?limit=100&before={data[-1]['id']}")
            if len(data) > 0:
                for msg in data:
                    messages.append(Message(msg, self.bot, self.http))
            else:
                break

        return messages

    async def purge(self, amount: int = None):
        messages = await self.history()
        msgs = []
        for msg in messages:
            if str(msg.author.id) == str(self.bot.user.id):
                msgs.append(msg)

        if amount != None:
            for i in range(0, len(msgs[:amount]), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[:amount][i:i + 3]))
        else:
            for i in range(0, len(msgs), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[i:i + 3]))

    async def spam(self, amount: int, content: str, tts=False):
        amount: list[int] = [i + 1 for i in range(amount)]
        for i in range(0, len(amount), 3):
            await asyncio.gather(
                *(asyncio.create_task(self.send(tts=tts, content=content)) for amoun in amount[i:i + 3]))

    async def send(self, content=None, tts=False):
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/@me/{self.id}"},
                                json={"content": content, "tts": tts})

    async def reply(self, message, content=None, tts=False):
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/@me/{self.id}"},
                                json={"content": content, "tts": tts,
                                      "message_reference": {"channel_id": f"{self.id}", "message_id": f"{message.id}"},
                                      "allowed_mentions": {"parse": ["users", "roles", "everyone"],
                                                           "replied_user": False}})


class GroupChannel:
    """Group Channel Object
    """

    def __init__(self, data, bot, http) -> None:
        self.recipients = []
        self.http = http
        self.bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def _update(self, data):
        for user in data.get("recipients"):
            self.recipients.append(User(user, self.bot, self.http))
        self.name = data.get("name")
        self.owner_id = data.get("owner_id")
        self.last_message_id = data.get("last_message_id")
        self.id = data.get("id")
        self.flags = data.get("flags")
        self.icon = data.get("icon")

    async def delete(self):
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}?silent=true")
        del self

    async def history(self):
        messages = []
        data = await self.http.request(method="get", endpoint=f"/channels/{self.id}/messages?limit=100")
        for msg in data:
            messages.append(Message(msg, self.bot, self.http))
        while True:
            data = await self.http.request(method="get",
                                           endpoint=f"/channels/{self.id}/messages?limit=100&before={data[-1]['id']}")
            if len(data) > 0:
                for msg in data:
                    messages.append(Message(msg, self.bot, self.http))
            else:
                break

        return messages

    async def purge(self, amount: int = None):
        messages = await self.history()
        msgs = []
        for msg in messages:
            if str(msg.author.id) == str(self.bot.user.id):
                msgs.append(msg)

        if amount != None:
            for i in range(0, len(msgs[:amount]), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[:amount][i:i + 3]))
        else:
            for i in range(0, len(msgs), 3):
                await asyncio.gather(*(asyncio.create_task(message.delete()) for message in msgs[i:i + 3]))

    async def spam(self, amount: int, content: str, tts=False):
        amount: list[int] = [i + 1 for i in range(amount)]
        for i in range(0, len(amount), 3):
            await asyncio.gather(
                *(asyncio.create_task(self.send(tts=tts, content=content)) for amoun in amount[i:i + 3]))

    async def send(self, content=None, tts=False):
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/@me/{self.id}"},
                                json={"content": content, "tts": tts})

    async def reply(self, message, content=None, tts=False):
        await self.http.request(method="post", endpoint=f"/channels/{self.id}/messages", headers={"origin": "https://discord.com", "referer": f"https://discord.com/channels/@me/{self.id}"},
                                json={"content": content, "tts": tts,
                                      "message_reference": {"channel_id": f"{self.id}", "message_id": f"{message.id}"},
                                      "allowed_mentions": {"parse": ["users", "roles", "everyone"],
                                                           "replied_user": False}})





class Emoji:
    """Emoji Object
    """
    def __init__(self, data, bot, http) -> None:
        self.bot = bot
        self.http = http
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def _update(self, data):
        self.name = data.get("name")
        self.id = data.get("id")
        self.roles = data.get("roles")
        self.managed = data.get("managed")
        self.available = data.get("available")
        self.animated = data.get("animated")
        self.guild_id = data.get("guild_id")


    async def delete(self):
        await self.http.request(method="delete", endpoint=f"/guilds/{self.guild_id}/emojis/{self.id}")
        del self



from time import perf_counter
from ..models.role import Role
from ..models import User, Client, Guild, TextChannel, VoiceChannel, DMChannel, GroupChannel, Message
from aioconsole import aprint


class EventHandler:
    '''
    Used to handle discord events
    '''
    def __init__(self, bot, http):
        self._events = {}
        self.http    = http
        self.bot     = bot

    async def handle_ready(self, data, user: Client, http):
        '''Handles the ready event, what is executed when it appears
        '''
        self.user = user
        for relationship in data.get('relationships'):
            if relationship.get('type') == 1:
                self.user.friends.append(User(relationship['user'], self.bot, http))

        for channel in data.get('private_channels'):
            if channel.get('type') == 1: self.user.private_channels.append(DMChannel(channel,self.bot, http))
            if channel.get('type') == 3: self.user.private_channels.append(GroupChannel(channel, self.bot, http))

        for guild in data.get('guilds'):
            await self.handle_guild_create(guild, self.user, http)

        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit('ready', perf_counter() - self.bot.t1)

    async def handle_guild_create(self, data, user: Client, http):
        '''
        Handles what happens when a guild is created
        '''
        self.user = user
        guild = Guild(data, self.bot, http)
        self.user.guilds.append(guild)

        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit('guild_create', guild)

    async def handle_message_create(self, data, user: Client, http):
        '''Handles what happens when a message is created
        '''
        self.user = user
        message = Message(data, self.bot, http)
        self.user.messages.append(message)

        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit('message', message)
        if message.author.id == self.bot.user.id:
            for prefix in self.bot.prefixes:
                # Attempts to invoke the command if has prefix and from the user
                if message.content.startswith(prefix): await self.bot.process_commands(message)

    async def handle_message_delete(self, data, user: Client, http):
        '''Handles what happens when a message is created. Disclaimer: Only guild id, message id and channel id will be present if the message is not in bots cache.
        '''
        self.user = user
        id =data.get('id')
        for message in self.user.messages:
            if message.id == id:
                await self.bot.emit('message_delete', message)
                self.user.deleted_messages.append(message)
                self.user.messages.remove(message)
        else:
            class deleted_message:
                def __init__(self, data) -> None:
                    self.tts                = data.get('tts')
                    self.references_message = data.get('referenced_message')
                    self.mentions           = data.get('mentions')
                    self.author             = None
                    self.id                 = data.get('id')
                    self.flags              = data.get('flags')
                    self.embeds             = data.get('embeds')
                    self.content            = data.get('content')
                    self.components         = data.get('components')
                    self.channel_id         = data.get('channel_id')
                    self.attachments        = data.get('attachments')
                    self.guild_id           = data.get('guild_id')
                    self.channel            = None
                    self.guild              = None

            message = deleted_message(data)
            await self.bot.emit('message_delete', message)

    async def handle_channel_create(self, channel, user: Client, http):
        '''Handles what happens when a channel is created
        '''
        self.user = user
        if channel.get('type') == 0:
            id = channel.get('guild_id')
            for guild in self.user.guilds:
                if guild.id == id:
                    channel = TextChannel(channel, self.bot, self.http)
                    guild.channels.append(channel)

        elif channel.get('type') == 1: self.user.private_channels.append(DMChannel(channel, self.bot, http))

        elif channel.get('type') == 2:
            id = channel.get('guild_id')
            for guild in self.user.guilds:
                if guild.id == id:
                    channel = VoiceChannel(channel, self.bot, self.http)
                    guild.channels.append(channel)

        elif channel.get('type') == 3: self.user.private_channels.append(GroupChannel(channel, self.bot, http))

        else:
            id = channel.get('guild_id')
            for guild in self.user.guilds:
                if guild.id == id:
                    channel = TextChannel(channel, self.bot, self.http)
                    guild.channels.append(channel)

        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit('channel_create', channel)

    async def handle_guild_member_list_update(self, data, user: Client, http):
        '''Handles what happens when member chunk payload is sent via gateway
        '''
        members   = []
        self.user = user

        try:
            for item in data['ops'][0]['items']:
                try: members.append(User(item['member']['user'], self.bot, http))
                except: continue

            await self.bot.emit('member_chunk', members)
        except: pass



    async def handle_channel_delete(self, data, user: Client, http):
        '''Handles what happens when a channel is deleted
        '''
        self.user = user
        id = data.get('id')
        for channel in self.user.private_channels:
            if channel.id == id:
                await self.bot.emit('channel_delete', channel)
                self.user.private_channels.remove(channel)
                return

        guild_id = data.get('guild_id')
        if guild_id != None:
            for guild in self.user.guilds:
                if guild_id == guild.id:
                    for channel in guild.channels:
                        if channel.id == id:
                            await self.bot.emit('channel_delete', channel)
                            guild.channels.remove(channel)
                            return
        else:
            for guild in self.user.guilds:
                for channel in guild.channels:
                    if channel.id == id:
                        await self.bot.emit('channel_delete', channel)
                        guild.channels.remove(channel)
                        return


    async def handle_guild_role_create(self, role, user: Client, http):
        '''Handles what happens when a role is created
        '''
        self.user = user

        for guild in self.user.guilds:
            if role.get('guild_id') == guild.id:
                role = Role(role, self.http, guild_id=guild.id)
                guild.roles.append(role)

        await self.bot.emit('role_create', role)

    async def handle_guild_role_delete(self, role, user: Client, http):
        '''Handles what happens when a role is deleted
        '''
        self.user = user

        for guild in self.user.guilds:
            if role.get('guild_id') == guild.id:
                for role in guild.roles:
                    if role.get('id') == role.id:
                        await self.bot.emit('role_delete', role)
                        guild.roles.remove(role)
                        return















class Client:
    """Client Object
    """
    def __init__(self, UserPayload: dict) -> None:

        self.guilds = []
        self.private_channels = []
        self.friends = []
        self.messages = []
        self.deleted_messages = []
        self._update(UserPayload)

    def __str__(self):
        return f"""{self.name}#{self.discriminator}"""


    def _update(self, data):
        self.name = data.get("username")
        self.id = data.get("id")
        self.discriminator = data.get("discriminator")
        self.avatar = data.get("avatar")
        self.banner = data.get("banner")
        self.bio = data.get("bio")
        self.email = data.get("email")
        self.phone = data.get("phone")
        self.accent_colour = data.get('accent_color')
        self.public_flags = data.get('public_flags')
        self.bot = data.get('bot')
        self.system = data.get('system')





import asyncio
import websockets
from aioconsole import aprint
import json
import time
import zlib
from .events import EventHandler
from .errors import ReconnectWebsocket
from selfcord.models.client import Client

class Activity:

    @staticmethod
    def Game(name, details: str="", buttons: dict={}):
        type = 0
        button_urls = [button for button in buttons.values()]
        buttons: list = [button for button in buttons.keys()]

        if len(button_urls) == 0 or len(buttons) == 0:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "created_at": int(time.time()),
            }
        else:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time())
            }


        return payload

    @staticmethod
    def Stream(name, details: str="", url: str="https://www.youtube.com/watch?v=CyIrJVp-sH8",buttons: dict={}):
        type = 1
        button_urls = [button for button in buttons.values()]
        buttons: list = [button for button in buttons.keys()]

        if len(button_urls) == 0 or len(buttons) == 0:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "created_at": int(time.time()),
                "url": url
            }
        else:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "url": url,
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time())
            }

        return payload

    @staticmethod
    def Listen(name, details: str="", buttons: dict={}):
        type = 2
        button_urls = [button for button in buttons.values()]
        buttons: list = [button for button in buttons.keys()]

        if len(button_urls) == 0 or len(buttons) == 0:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "created_at": int(time.time()),
            }
        else:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time())
            }


        return payload

    @staticmethod
    def Watch(name, details: str="", buttons: dict={}):
        type = 3

        button_urls = [button for button in buttons.values()]
        buttons: list = [button for button in buttons.keys()]

        if len(button_urls) == 0 or len(buttons) == 0:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "created_at": int(time.time()),
            }
        else:
            payload = {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time())
            }


        return payload



class gateway:
    ''' OP CODES '''
    DISPATCH           = 0
    HEARTBEAT          = 1
    IDENTIFY           = 2
    PRESENCE           = 3
    VOICE_STATE        = 4
    VOICE_PING         = 5
    RESUME             = 6
    RECONNECT          = 7
    REQUEST_MEMBERS    = 8
    INVALIDATE_SESSION = 9
    HELLO              = 10
    HEARTBEAT_ACK      = 11
    GUILD_SYNC         = 12

    def __init__(self, http, show_heartbeat=False):
        self.show_heartbeat = show_heartbeat
        self.http = http
        self.zlib = zlib.decompressobj()
        self.zlib_suffix = b'\x00\x00\xff\xff'
        self.last_ack = time.perf_counter()
        self.last_send = time.perf_counter()
        self.latency = float('inf')
        self.alive = False

    async def recv_msg(self):
        '''
        Receives Message from gateway, encodes as json and does things depending on op code

        '''
        item = await self.ws.recv()
        buffer = bytearray()
        buffer.extend(item)
        if len(item) < 4 or item[-4:] != self.zlib_suffix: return

        if item:
            item    = self.zlib.decompress(item)
            item    = json.loads(item) # Get json message from gateway

            op      = item.get('op') # Op code
            data    = item.get('d') # Data
            event   = item.get('t') # The event


            if  op == self.RECONNECT:
                await self.close()
                raise ReconnectWebsocket('Connection was closed.')

            elif op == self.INVALIDATE_SESSION:
                if data:
                    await self.close()
                    raise ReconnectWebsocket('Connection was closed.')

            elif op == self.HELLO:
                # Begins heartbeat and sends identify if this op is received
                interval = data['heartbeat_interval'] / 1000.0
                await self.identify()
                asyncio.create_task(self.heartbeat(interval))

            elif op == self.HEARTBEAT_ACK:
                await self.heartbeat_ack()

            elif op == self.DISPATCH:
                # If op is 0 it signifies a regular gateway event
                # These events are discord events like message_create, role_create whatever.

                handle = f'handle_{event.lower()}'

                if hasattr(self.handler, handle): # If the event handler exists, so e.g handle_ready
                    method = getattr(self.handler,handle)

                    val = await asyncio.gather(asyncio.create_task(method(data, self.user, self.http)), return_exceptions=True) # A background task is created to run the handler
                    for item in val:
                        if item == None: break
                        else: await self.bot.emit('error', item)

                    # asyncio.create_task(method(data, self.user, self.http))
                # Handlers are all situated in events.py


    def roundup(self, n):
        import math
        return int(math.ceil(n / 100.0)) * 100

    def chunks(self, lst, n):
        for i in range(0, len(lst), 1):
            if len(lst[:i+1]) > 3:
                for i in range(i, len(lst), n): yield lst[i:i + n]
                break

            yield lst[:i+1]

    async def change_presence(self, status: str, afk: bool=False, activity: dict= Activity.Game("Selfcord", "Greatest wrapper" )):
        """Change the clients current presence

        Args:
            status (str): online, offline or dnd
            afk (bool): Whether client is set as AFK
            activity (Activity): Activity object
        """
        payload = {
            "op": 3,
            "d" : {
                "since": time.time(),
                "activities": [activity],
                "status": status.lower(),
                "afk": afk
            },
        }

        await self.send_json(payload)



    async def lazy_chunk(self, guild_id: str, channel_id: str, amount: int):
        '''Sends lazy guild request to gather current online members

        Args:
            guild_id (str): The guild id specified
            channel_id (str): The channel id specified
        '''

        ranges = []

        for i in range(0, amount, 100):
            ranges.append([i, self.roundup(i + (amount - i)) - 1]) if i + 99 > amount else ranges.append([i, i+99])

        for item in self.chunks(ranges, 3):
            payload = {
                'op': 14,
                'd': {
                    'guild_id': guild_id,
                    'typing': True,
                    'channels': {channel_id:item}
                }
            }

            await self.send_json(payload)


    async def send_json(self, payload: dict):
        '''Send json to the gateway

        Args:
            payload (dict): Valid payload to send to the gateway
        '''
        await self.ws.send(json.dumps(payload))

    async def connect(self):
        '''Connect to discord gateway
        '''
        self.ws = await websockets.connect('wss://gateway.discord.gg/?encoding=json&v=9&compress=zlib-stream', origin='https://discord.com')
        self.alive = True

    async def close(self):
        '''Close the connection to discord gateway
        '''
        self.alive= False
        await self.ws.close()

    async def identify(self):
        '''Identify to gateway, uses amazing mobile client spoof
        '''
        payload = {
            'op': 2,
            'd': {
                'token': self.token,
                'properties': {
                    '$os': 'android',
                    '$browser': 'Discord Android',
                    '$device': 'Discord Android',
                    '$referrer': '',
                    '$referring_domain': ''
                },
            }
        }
        await self.send_json(payload)



    async def heartbeat(self, interval):
        '''Heartbeat for gateway to maintain connection

        Args:
            interval (int): Interval between sends
        '''
        await aprint(f'Hearbeat loop has began with the interval of {interval} seconds!')
        heartbeatJSON = {
            'op': 1,
            'd': time.time()
        }
        while True:
            await asyncio.sleep(interval)
            await self.send_json(heartbeatJSON)
            self.last_send = time.perf_counter()
            if self.show_heartbeat:
                await aprint('Sent Beat')

    async def heartbeat_ack(self):
        '''Whenever heartbeat ack is sent, logs the time between last send of heartbeat json and receive of the ack
        '''
        self.last_ack = time.perf_counter()
        self.latency = self.last_ack - self.last_send

    async def start(self, token: str, user: Client, bot):
        '''Start discord gateway connection

        Args:
            token (str): User token
            user (Client): User client
            bot (_type_): Bot class
        '''
        self.handler = EventHandler(bot, self.http)
        self.bot = bot

        await self.bot.inbuilt_commands() # In built commands very cool

        self.user = user
        self.token = token

        await self.connect()
        while self.alive:
            try: await self.recv_msg()
            except KeyboardInterrupt:
                await aprint('Shutting down...')
                await self.close()
            except Exception as e:
                await self.bot.emit('error', e)
                await self.close()


















from .user import User
from .channel import TextChannel, VoiceChannel, Category
from .role import Role
from .emoji import Emoji
from itertools import zip_longest
from aiohttp import ClientSession
from base64 import b64encode
import random
class Guild:
    """Guild Object
    """
    TEXTCHANNEL = 0
    VOICECHANNEL = 2
    CATEGORY = 4
    GUILD_ANNOUNCEMENT = 5
    ANNOUNCEMENT_THREAD = 10
    PUBLIC_THREAD = 11
    PRIVATE_THREAD = 12
    GUILD_STAGE_VOICE = 13
    GUILD_DIRECTORY = 14
    GUILD_FORUM = 15

    def __init__(self, data, bot, http) -> None:
        self.roles = []
        self.emojis = []
        self.members = []
        self.channels = []
        self.http = http
        self.bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def _update(self, data):
        self.id = data.get('id')
        self.name = data.get('name')
        self.icon = data.get('icon')
        self.region = data.get('region')
        self.splash = data.get('splash')
        self.mfa_level = data.get('mfa_level')
        self.features = data.get('features')
        self.member_count = data.get("member_count")
        self.unavailable = data.get('unavailable')
        self.verification_level = data.get('verification_level')
        self.explicit_content_filter = data.get('explicit_content_filter')
        self.owner_id = data.get('owner_id')

        for (member, channel, role, emoji) in zip_longest(data.get('members'), data.get("channels"), data.get("roles"), data.get("emojis")):
            if member != None:
                user = User(member, self.bot, self.http)
                self.members.append(user)

            if channel != None:
                type = channel.get("type")
                if type == self.TEXTCHANNEL:
                    channel = TextChannel(channel, self.bot, self.http)
                    self.channels.append(channel)
                elif type == self.VOICECHANNEL:
                    channel = VoiceChannel(channel, self.bot, self.http)
                    self.channels.append(channel)
                elif type == self.CATEGORY:
                    channel = Category(channel, self.bot, self.http)
                    self.channels.append(channel)
                else:
                    channel = TextChannel(channel, self.bot, self.http)
                    self.channels.append(channel)

            if role != None:
                role = Role(role, self.http, guild_id = self.id)
                self.roles.append(role)

            if emoji != None:
                emoji = Emoji(emoji, self.bot, self.http)
                self.emojis.append(emoji)

    async def txt_channel_create(self, name, parent_id=None):
        payload = {"name": name}
        payload.update({"permission_overwrites": []})
        payload.update({"type": 0})
        if parent_id:
            payload.update({"parent_id": parent_id})

        await self.http.request(method="post", endpoint=f"/guilds/{self.id}/channels", json=payload)

    async def vc_channel_create(self, name):
        await self.http.request(method="post", endpoint=f"/guilds/{self.id}/channels", json={"name": f"{name}", "permission_overwrites": [], "type": 2})

    async def role_create(self, name):
        await self.http.request(method = "post", endpoint = f"/guilds/{self.id}/roles", json = {"name": f"{name}"})

    async def category_channel_create(self, name):
        await self.http.request(method = "post", endpoint = f"/guilds/{self.id}/channels", json={"name": f"{name}", "permission_overwrites": [], "type": 4})

    async def emoji_create(self, name: str, image_url: str):
        image = await self.http.encode_image(image_url)
        await self.http.request(method = "post", endpoint = f"/guilds/{self.id}/emojis", json= {"name": f"{name}", "image": image})

    async def get_members(self, channel_id: str):
        await self.bot.gateway.lazy_chunk(self.id, channel_id, self.member_count)

    async def edit(self, name: str=None, icon_url: str=None, banner_url: str=None, description: str=None):
        fields = {}
        if name != None:
            fields['name'] = name

        if description != None:
            fields['description'] = description

        if icon_url != None:
            data = await self.http.encode_image(icon_url)
            fields['icon'] = data

        if banner_url != None:
            data = await self.http.encode_image(banner_url)
            fields['banner'] = data


        await self.http.request(method = "patch", endpoint = f"/guilds/{self.id}", headers={"origin":"https://discord.com", "referer":f"https://discord.com/channels/{self.id}/{random.choice(self.channels)}"},json=fields)

    async def delete(self):
        await self.http.request(method = "delete", endpoint = f"/guilds/{self.id}")




from .user import User
import asyncio

class Message:
    """Message Object
    """
    def __init__(self, data, bot, http) -> None:
        self.bot = bot
        self.channel = None
        self.guild = None
        self.http = http
        self._update(data)

    def __str__(self) -> str:
        return f"{self.content}"

    def _update(self, data):
        self.tts = data.get("tts")
        self.references_message = data.get("referenced_message")
        self.mentions = data.get("mentions")
        self.author = User(data.get("author"), self.bot, self.http)
        self.id = data.get("id")
        self.flags = data.get("flags")
        self.embeds = data.get("embeds")
        self.content = data.get("content")
        self.components = data.get("components")

        self.channel_id = data.get("channel_id")

        self.attachments = data.get("attachments")
        self.guild_id = data.get("guild_id")

        self.channel = self.bot.get_channel(self.channel_id)

        self.guild = self.bot.get_guild(self.guild_id)

    async def delete(self):
        await self.http.request(method="delete", endpoint=f"/channels/{self.channel_id}/messages/{self.id}")












from .permission import Permission

class Role:
    """Role Object
    """
    def __init__(self, data: dict, http, **kwargs) -> None:
        self.guild_id = kwargs.get("guild_id")
        self.http = http
        self._update(data)

    def _update(self, data):
        role = data.get("role")

        self.id = data.get("id") if role==None else role.get("id")
        self.permissions = Permission(int(data.get("permissions"))).permissions if role==None else Permission(int(role.get("permissions"))).permissions
        self.name = data.get("name") if role==None else role.get("name")
        self.mentionable = data.get("mentionable") if role==None else role.get("mentionable")
        self.managed = data.get("managed") if role==None else role.get("managed")
        self.icon = data.get("icon") if role==None else role.get("icon")
        self.flags = data.get("flags") if role==None else role.get("flags")

        self.color = data.get("color") if role==None else role.get("color")
        self.hoist = data.get("hoist") if role==None else role.get("hoist")
        if self.guild_id is None:
            self.guild_id = data.get("guild_id")



    def __str__(self) -> str:
        return f"{self.name}"

    async def delete(self):
        self.http.request(method="delete", endpoint=f"/guilds/{self.guild_id}/roles/{self.id}")
        del self


class DiscordException(Exception):
    pass

class LoginFailure(DiscordException):
    def __init__(self, message: dict, status: int) -> None:
        self.message = message
        self.status = status

class ReconnectWebsocket(DiscordException):
    def __init__(self, message: str) -> None: self.message = message

class RuntimeError(DiscordException):
    def __init__(self, message: str) -> None: self.message = message

class Funnu(DiscordException):
    def __init__(self, *args: object) -> None: super().__init__(*args)


from .command import CommandCollection


class Extension:
    """Extension object pretty much
    """
    def __init__(self, **kwargs):
        self.name: str | None = kwargs.get("name")
        self.description: str | None = kwargs.get('description')
        self.ext = kwargs.get("ext")
        self.commands: CommandCollection | None = kwargs.get("commands")

class Extender:
    def __init__(self, bot) -> None:
        self.commands = CommandCollection(bot)
    class Extension:
        def __init_subclass__(cls, name=None, description="") -> None:
            super().__init_subclass__()
            cls.name = name
            cls.description = description




class ExtensionCollection:
    """Extension collection, where extensions are stored into
    """
    def __init__(self, bot):
        self.bot = bot
        self.extensions = {}

    def __iter__(self):
        for cmd in self.extensions.values():
            yield cmd

    def _is_already_registered(self, ext):
        for extension in self.extensions.values():
            if ext.name == extension:
                return True

    def add(self, ext):
        if not isinstance(ext, Extension):
            raise ValueError('cmd must be a subclass of Command')
        if self._is_already_registered(ext):
            raise ValueError('A name or alias is already registered')
        self.extensions[ext.name] = ext

    def get(self, alias, prefix=''):
        try:
            return self.extensions[alias]
        except KeyError:
            pass
        for command in self.extensions:
            if command.name in command.aliases:
                return command




from base64 import b64encode
import datetime

class User:
    """User Object
    """
    def __init__(self, UserPayload: dict, bot, http) -> None:
        self.bot = bot
        self.http = http
        self._update(UserPayload)

    def __str__(self):
        return f"""{self.name}#{self.discriminator}"""

    @property
    def created_at(self):
        return datetime.datetime.utcfromtimestamp(((int(self.id) >> 22) + 1420070400000) / 1000)

    @property
    def b64token(self):
        return str(b64encode(self.id.encode("utf-8")), "utf-8")

    def _update(self, data):
        self.name = data.get("username")
        self.id = data.get("id")
        self.discriminator = data.get("discriminator")
        self.avatar = data.get("avatar")
        self.banner = data.get("banner")
        self.accent_colour = data.get('accent_color')
        self.public_flags = data.get('public_flags')
        self.bot = data.get('bot')
        self.avatar_url = f"https://cdn.discordapp.com/avatars/{self.id}/{self.avatar}.png?size=4096" if self.avatar!=None else None
        self.banner_url = f"https://cdn.discordapp.com/banners/{self.id}/{self.banner}.png?size=1024" if self.banner!=None else None
        self.system = data.get('system')


    async def create_dm(self):
        await self.http.request(method="post", endpoint="/users/@me/channels", json={"recipients": [self.id]})

    



from setuptools import setup, find_packages

if __name__ == "__main__":
    setup(
        name="selfcord.py",
        packages=find_packages(include=['selfcord', 'selfcord.api', 'selfcord.utils', 'selfcord.models']),
        version="0.0.2",
        description="A Discord API wrapper designed for selfbots!",
        author="Shell of OMEGA",
        license="MIT",
        install_requires=["aiohttp==3.7.4.post0","aioconsole==0.3.3", "websockets==10.1"],
        setup_requires=['pytest-runner'],
        tests_require=['pytest'],
        test_suite='tests',
    )


import asyncio
import json
from .api import gateway, http, Activity
import inspect
from .models import Client, TextChannel, GroupChannel, DMChannel, VoiceChannel, Guild, User
from collections import defaultdict
from aioconsole import aprint, aexec
import time
from .utils import Command, CommandCollection, Context, ExtensionCollection, Extension
import random
import contextlib
from traceback import format_exception
import io
import importlib


class Bot:
    def __init__(self, show_beat: bool = False, prefixes: list = ["s!"]) -> None:
        self.show_beat = show_beat
        self.token = None
        self.http = http()
        self.t1 = time.perf_counter()
        self.gateway = gateway(self.http, self.show_beat)
        self._events = defaultdict(list)
        self.commands = CommandCollection(self)
        self.extensions = ExtensionCollection(self)
        self.prefixes = prefixes if isinstance(prefixes, list) else [prefixes]
        self.extensions = {}
        self.user = None

    def run(self, token: str):
        """Used to start connection to gateway as well as gather user information

        Args:
            token (str): _description_
        """
        self.token = token

        async def runner():
            data = await self.http.static_login(token)
            self.user = Client(data)
            await self.gateway.start(token, self.user, self)

        try:
            asyncio.run(runner())
        except KeyboardInterrupt:
            pass

    @property
    def latency(self):
        """Latency of heartbeat ack, gateway latency essentially"""
        return self.gateway.latency

    # For events
    async def inbuilt_commands(self):
        """
        I call this on bot initialisation, it's the inbuilt help command
        """

        @self.cmd("The help command!", aliases=["test"])
        async def help(ctx):
            msg = f"```diff\n"
            msg += f"+ {self.user} selfbot\n+ Prefixes:   {self.prefixes}\n\n"
            msg += f"- Commands\n"
            for command in self.commands:
                msg += f"- {command.name}:    {command.description}\n"

                if len(msg) > 1980:
                    msg += f"```"
            msg += f"```"
            await ctx.reply(f"{msg}")

        def clean_code(content):
            if content.startswith("```") and content.endswith("```"):
                return "\n".join(content.split("\n")[1:])[:-3]
            else:
                return content

        @self.cmd(description="Evaluates and runs code", aliases=['exec'])
        async def eval(ctx, *, code: str):
            code = clean_code(code)

            try:
                with contextlib.redirect_stdout(io.StringIO()) as f:
                    await aexec(code)
                    result = f"```{f.getvalue()}\n```"
            except Exception as e:
                error = "".join(format_exception(e, e, e.__traceback__))
                result = f"```\n{error}```"

            await ctx.reply(result)

    def on(self, event: str):
        """Decorator for events

        Args:
            event (str): The event to check for
        """

        def decorator(coro):
            if not inspect.iscoroutinefunction(coro):
                raise RuntimeWarning("Faulure")
            else:
                self._events[event].append(coro)

                def wrapper(*args, **kwargs):
                    result = self._events[event].append(coro)
                    return result

                return wrapper

        return decorator

    async def emit(self, event, *args, **kwargs):
        """Used to essentially push values to the decorator

        Args:
            event (str): The event name
        """
        on_event = "on_{}".format(event)
        try:
            if hasattr(self, on_event):
                await getattr(self, on_event)(*args, **kwargs)
            if event in self._events:
                for callback in self._events[event]:
                    asyncio.create_task(callback(*args, **kwargs))
        except Exception as e:
            raise RuntimeError("Failure to emit")

    def cmd(self, description="", aliases=[]):
        """Decorator to add commands for the bot

        Args:
            description (str, optional): Description of command. Defaults to "".
            aliases (list, optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]

        def decorator(coro):
            name = coro.__name__
            if not inspect.iscoroutinefunction(coro):
                raise RuntimeWarning("Not an async function!")
            else:
                cmd = Command(name=name, description=description, aliases=aliases, func=coro)
                self.commands.add(cmd)
            return cmd

        return decorator

    def add_cmd(self, coro, description="", aliases=[]):
        """
        Function to add commands manually without decorator

        Args:
            coro (coroutine): The function to add
            description (str, optional): Description of command. Defaults to "".
            aliases (list, optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]
        name = coro.__name__
        if not inspect.iscoroutinefunction(coro):
            raise RuntimeWarning("Not an async function!")
        else:
            cmd = Command(name=name, description=description, aliases=aliases, func=coro)
            self.commands.add(cmd)

    async def process_commands(self, msg):
        """
        What is called in order to actually get command input and run commands

        Args:
            msg (str): The message containing command
        """
        context = Context(self, msg, self.http)

        asyncio.create_task(context.invoke())

    async def load_extension(self, name: str):
        """
        Load various extensions/plugins/cogs if any.

        Args:
            name (str): Name of the extension to load

        Raises:
            ModuleNotFoundError: If you suck and don't know the name of what you want to load
        """
        try:
            name = importlib.util.resolve_name(name, None)
        except Exception as e:
            raise ModuleNotFoundError(f"{name} does not exist")

        spec = importlib.util.find_spec(name)

        lib = importlib.util.module_from_spec(spec)

        try:
            spec.loader.exec_module(lib)
        except Exception as e:
            raise ModuleNotFoundError(f"Spec could not be loaded")
        try:
            ext = getattr(lib, 'Ext')
        except Exception as e:
            raise ModuleNotFoundError(f"Extension does not exist")


        self.extensions.add(Extension(name=ext.name, description=ext.description, ext=ext))

    def add_ext(self):
        """
        To be built later.
        Note: Need contributors.
        """
        print("okaowdkoawkdoakwd")



    def get_channel(self, channel_id: str):
        """
        Function to help retrieve channel from bot cache

        Args:
            channel_id (str): The channel id to search for

        Returns:
            Channel: The Channel object
        """
        for channel in self.user.private_channels:
            if channel_id == channel.id:
                return channel
        for guild in self.user.guilds:
            for channel in guild.channels:
                if channel_id == channel.id:
                    return channel

    def get_guild(self, guild_id: str):
        """
        Function to help retrieve guild from bot cache

        Args:
            guild_id (str): The guild id to search for

        Returns:
            Guild: The Guild object
        """
        for guild in self.user.guilds:
            if guild.id == guild_id:
                return guild

    async def get_user(self, user_id: str):
        """
        Function to retrieve user data. Probably need to be friends with them to retrieve the details.

        Args:
            user_id (Str): ID of the other user.

        Returns:
            User: The User object
        """
        data = await self.http.request(method="get", endpoint=f"/users/{user_id}")
        return User(data)

    async def add_friend(self, user_id: str):
        """
        Function to add random user as a friend.

        Args:
            user_id (str): ID of the possible random user.

        Returns:
            No return value.
        """
        await self.http.request(method="put", endpoint=f"/users/@me/relationships/{user_id}",
                                headers={"origin": "https://discord.com",
                                         "referer": f"https://discord.com/channels/@me/{random.choice(self.user.private_channels).id}"},
                                json={})

    async def edit_profile(self, bio: str = None, accent: int = None):
        """

        """
        fields = {}
        if bio != None:
            fields['bio'] = bio
        if accent != None:
            fields['accent'] = accent
        await self.http.request(method="patch", endpoint=f"/users/@me/profile", json=fields)

    async def change_pfp(self, avatar_url=None):
        """Disclaimer: This may phone lock your account :(

        Args:
            avatar_url (str): URL of image

        Raises:
            TypeError: URL not specified
        """
        if avatar_url != None:
            image = await self.http.encode_image(avatar_url)
            await self.http.request(method="patch", endpoint="/users/@me", headers={"origin": "https://discord.com",
                                                                                    "referer": "https://discord.com/channels/@me"},
                                    json={'avatar': image})
        else:
            raise TypeError("Avatar url not specified")

    async def create_dm(self, recipient_id: int):
        """
        Function to create new DM Channel with other user. Can be used with bots too.

        Args:
            recipient_id (snowflake): ID of recipient - Has to be user or bot ID

        Raises:
            TypeError: Recipient ID not specified

        Returns:
            DMChannel object
        """
        if recipient_id is not None:
            data = await self.http.request(method="post", endpoint="/users/@me/channels",
                                           json={"recipient_id": recipient_id})
            return DMChannel(data, bot=self, http=self.http)
        else:
            raise TypeError("Recipient ID not specified")

    async def change_hypesquad(self, house: str):
        if house.lower() == "bravery":
            await self.http.request(method="post", endpoint = "/hypesquad/online", json = {"house_id": 1})
        if house.lower() == "brilliance":
            await self.http.request(method="post", endpoint = "/hypesquad/online", json = {"house_id": 2})
        if house.lower() == "balance":
            await self.http.request(method="post", endpoint = "/hypesquad/online", json = {"house_id": 3})

    async def change_presence(self, status: str, afk: bool=False, activity: dict=Activity.Game("test", "testing")):
        await self.gateway.change_presence(status, afk, activity=activity)








class Member:
    """Member Object
    """
    def __init__(self, UserPayload: dict) -> None:
        self.roles = []

        self._update(UserPayload)

    def __str__(self):
        return f"""{self.name}#{self.discriminator}"""


    def _update(self, data):
        user = data.get("user")
        self.name = user.get("username")
        self.id = user.get("id")
        self.discriminator = user.get("discriminator")
        self._avatar = user.get("avatar")
        self._banner = user.get("banner")
        self._accent_colour = user.get('accent_color')
        self._public_flags = user.get('public_flags')
        self.bot = user.get('bot')
        self.joined_at = data.get('joined_at')
        self.nick = data.get("nick")
        self.system = data.get('system')




import inspect
import re







class Command:
    """Command Object pretty much
    """
    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.aliases = [self.name] + kwargs.get('aliases', [])
        self.description = kwargs.get('description')
        self.func = kwargs.get("func")
        self.check = inspect.signature(self.func).return_annotation
        self.signature = inspect.signature(self.func).parameters.items()

class CommandCollection:
    """Commands collection, where commands are stored into
    """
    def __init__(self, bot):
        self.bot = bot
        self.commands = {}

    def __len__(self):
        return len(self.commands)
    def __iter__(self):
        for cmd in self.commands.values():
            yield cmd

    def _is_already_registered(self, cmd):
        for command in self.commands.values():
            for alias in cmd.aliases:
                if alias in command.aliases:
                    return True

    def add(self, cmd):
        if not isinstance(cmd, Command):
            raise ValueError('cmd must be a subclass of Command')
        if self._is_already_registered(cmd):
            raise ValueError('A name or alias is already registered')
        self.commands[cmd.name] = cmd

    def get(self, alias, prefix=''):
        try:
            return self.commands[alias]
        except KeyError:
            pass
        for command in self.commands:
            if alias in command.aliases:
                return command

class Context:
    """Context related for commands, and invokation
    """
    def __init__(self, bot, message, http) -> None:
        self.bot = bot
        self.message = message
        self.http = http


    @property
    def author(self):
        return self.message.author

    @property
    def guild(self):
        return self.message.guild

    @property
    def channel(self):
        return self.message.channel

    @property
    def content(self):
        return self.message.content

    @property
    def command(self):
        if self.prefix is None:
            return None
        for command in self.bot.commands:
            for alias in command.aliases:
                if self.content.startswith(self.prefix + alias):
                    return command
        return None

    @property
    def alias(self):
        for command in self.bot.commands:
            for alias in command.aliases:

                if self.content.startswith(self.prefix + alias):
                    return alias
        return None

    @property
    def prefix(self):
        for prefix in self.bot.prefixes:
            if self.content.startswith(prefix):
                return prefix

    @property
    def command_content(self):
        """The content minus the prefix and command name, essentially the args

        Returns:
            str: String of content
        """
        if self.alias == None:
            return
        try:
            cut = len(self.prefix + self.alias)
            return self.content[cut:]
        except:
            return None

    def get_converter(self, param):
        if param.annotation is param.empty:
            return str
        if callable(param.annotation):
            return param.annotation
        else:
            raise ValueError('Parameter annotation must be callable')

    def convert(self, param, value):
        """Attempts to turn x value in y value, using get_converter func for the values

        Args:
            param (_type_): function parameter
            value (_type_): value in message

        Returns:
            Type[str]: The type of parameter
        """
        converter = self.get_converter(param)
        return converter(value)

    async def get_arguments(self):
        """Get arguments by checking function arguments and comparing to arguments in message.

        Returns:
            _type_: _description_
        """
        args = []
        kwargs = {}

        if self.command.signature is not None:
            signature = self.command.signature
        if self.command_content != "":
            splitted = self.command_content.split(" ")[1:]

            for index, item in enumerate(splitted):
                user_regex = re.findall(r"<@[0-9]{18,19}>", item)
                if len(user_regex) > 0:
                    x = re.findall(r"[0-9]{18,19}", item)
                    if len(x) > 0:
                        val = x[0]
                        splitted[index] = val
                break
        else:
            return args, kwargs


        for index, (name, param) in enumerate(signature):
            if index == 0:
                continue

            if param.kind is param.POSITIONAL_OR_KEYWORD:

                arg = self.convert(param, splitted.pop(0))

                args.append(arg)
            if param.kind is param.VAR_KEYWORD:

                for arg in splitted:
                    arg = self.convert(param, arg)
                    args.append(arg)


            if param.kind is param.KEYWORD_ONLY:

                arg = self.convert(param, ' '.join(splitted))
                kwargs[name] = arg


        for key in kwargs.copy():
            if not kwargs[key]:
                kwargs.pop(key)

        return args, kwargs




    async def invoke(self):
        """Used to actually run the command
        """
        if self.command is None:
            return
        if self.message.author.id != self.bot.user.id:
            return
        if self.command_content != None:
            args, kwargs = await self.get_arguments()
            func = self.command.func
            args.insert(0, self)


        await func(*args, **kwargs)



    async def reply(self, content: str, tts=False):
        """Helper function to reply to your own message containing the command

        Args:
            content (str): The message you would like to send
            tts (bool, optional): Whether message should be tts or not. Defaults to False.
        """
        await self.channel.reply(self.message, content, tts)

    async def send(self, content: str, tts=False):
        """Helper function to send message to the current channel

        Args:
            content (str): The message you would like to send
            tts (bool, optional): Whether message should be tts or not. Defaults to False.
        """
        await self.channel.send( content=content, tts=tts)

    async def spam(self, amount: int, content: str):
        """Helper function to spam messages in the current channel (uses asyncio.gather !!!!)

        Args:
            amount (int): Amount of messages to spam
            content (str): The message you would like to send
        """
        await self.channel.spam(amount, content)

    async def purge(self, amount: int=None):
        """Helper function to purge messages in the current channel, uses asyncio gather.

        Args:
            amount (int): The amount of messages to purge, defaults to All.
        """
        await self.channel.purge(amount)




